import re

from .message import msg
from .place import Place
from .player import Player
from .save import save_game, load_game
from .user_answer import get_answer_no_confirm, get_answer_confirm


def explore_place(place: Place, player: Player, all_places: dict) -> Place|None:
    msg.scroll_narrate(place.description)
    while True:
        command = input('> ')
        if command.lower() == 'status':
            player.status()
            continue
        elif command.lower().startswith('investigate'):
            target = command.replace('investigate', '').strip()
            for feature in place.features:
                if feature.name in target:
                    break
            else:
                msg.warning('\nInvestigate what, exactly?\n')
                continue
            msg.scroll_narrate(feature.interact_msg)
            if feature.danger:
                msg.scroll_danger(feature.danger_msg)
                player.wound(feature.danger)
        elif command.lower().startswith('take'):
            target = command.replace('take', '').strip()
            for feature in place.features:
                if feature.name in target:
                    break
            else:
                msg.warning('Take what now?')
                continue
            if not feature.takeable:
                msg.warning(f'That {feature.name} is not going to help you much. Try something else.')
                continue
            else:
                player.take(feature.takeable)
                place.features.remove(feature)
        elif command.lower().startswith('read'):
            target = command.replace('read', '').strip()
            player.read(target)
        elif command.lower().startswith('use'):
            tool_target = re.sub(r'use|on|the|in', '', command).split()
            tool = tool_target[0]
            target = tool_target[1]
            player.use(tool, place, target)
        elif command.lower().startswith('go'):
            target = re.sub(r'go|to|the|through|', '', command).strip()
            for node in place.nodes:
                if target in node.name:
                    break
            else:
                msg.warning("Go where now?")
                continue
            if node.accessible:
                msg.scroll_narrate(node.travel_msg)
                return all_places[node.place_name]
            else:
                msg.warning('That way is inaccessible to you right now.')
                continue
        elif command.lower() in ('backpack', 'inventory'):
            player.show_inventory()
        elif command.lower().startswith('eat'):
            target = command.replace('eat', '').strip()
            player.eat(target)
        elif command.lower() in ('quit', 'stop', 'exit', 'close'):
            msg.warning('Do you want to save first?')
            answer = get_answer_no_confirm(('yes', 'no'))
            if answer == 'yes':
                save_game(player, place, all_places)
            return
        elif command.lower() == 'save':
            save_game(player, place, all_places)

def new_game(new_game_msg: str, all_places: dict[str, Place]):
    msg.orange(new_game_msg)
    name = get_answer_confirm()
    player = Player(name)
    msg.dialogue(f'Hello, {player.name}. You can "continue" or ask for "help"')
    get_answer_no_confirm(('continue'))
    current_place = list(all_places.values())[0]
    return player, current_place, all_places

def main_loop(current_place: Place, player: Player, all_places: dict):
    while True:
        if isinstance(current_place, Place):
            current_place = explore_place(current_place, player, all_places)
        else:
            break
    msg.dialogue('\n[bold]Thanks for playing!\n')

def start(new_game_msg: str, all_places: dict[str, Place]):
    msg.dialogue('[u]Load[/u] game or start a [u]new[/u] game?\n')
    answer = get_answer_no_confirm(('load', 'new'))
    if answer == 'load':
        data = load_game()
        if None in (data):
            player, current_place, all_places = new_game(new_game_msg, all_places)
        else:
            player, current_place, all_places = data
    elif answer == 'new':
        player, current_place, all_places = new_game(new_game_msg, all_places)
    main_loop(current_place, player, all_places)



from .message import msg

def get_answer_confirm():
    while True:
        answer = input('> ')
        confirmation = input(f'Are you happy with "{answer}"?: ')
        if confirmation == 'yes':
            return answer
        else:
            msg.orange('Okay, try again.')
            continue

def get_answer_no_confirm(valid_responses: tuple[str]):
    while True:
        answer = input('> ')
        if answer.lower() == 'help':
            msg.help_text()
            continue
        if answer in valid_responses:
            return answer
        else:
            msg.plain(f"That wasn't one of the following valid responses: {valid_responses}")


from pathlib import Path
import pickle

from .message import msg
from .player import Player
from .place import Place
from .user_answer import get_answer_no_confirm

BASE_DIR = Path(__file__).parent.resolve()
OUTPUT_DIR = BASE_DIR / 'save_files'

def save_game(player: Player, current_place: Place, all_places: dict[str, Place]):
    OUTPUT_DIR.mkdir(exist_ok=True)
    output = OUTPUT_DIR / player.name
    data = (player, current_place, all_places)
    with open(output, 'wb') as f:
        pickle.dump(data, f)
    msg.success(f'Successfully saved game as {player.name}')
    

def load_game() -> tuple[Player, Place, dict[str, Place]]:
    save_dir = OUTPUT_DIR.glob('*')
    saved_games: list[Path] = []
    for f in save_dir:
        if f.is_file():
            saved_games.append(f)
    total_saved = len(saved_games)
    if total_saved == 0:
        msg.warning('Sorry, there were no saved game files found. Starting a new game...')
        return None, None, None
    number = ('are', 's') if total_saved >= 2 else ('is', '')
    msg.orange(f'There {number[0]} {total_saved} saved game{number[1]}. Enter a name to load it.')
    for game in saved_games:
        msg.plain(game.stem)
    game = get_answer_no_confirm([f.stem for f in saved_games])
    game_dir = OUTPUT_DIR / game
    with open(game_dir, 'rb') as f:
        data = pickle.load(f)
    msg.success(f'Loaded {data[0].name}')
    return data



from .message import msg
from .items import Consumable, Tool, Clue
from .place import Place


class Player:
    def __init__(self, name: str) -> None:
        self.score = 0
        self.name = name
        self.health = 100
        # self.strength = 1
        self.inventory: list[Tool|Consumable|Clue] = []
    
    def take(self, item: Consumable|Tool|Clue):
        self.inventory.append(item)
        msg.success(f'You put the {item.name} in your inventory.')

    def drop(self, name: str):
        new_inventory: list[Tool|Consumable|Clue] = []
        for item in self.inventory:
            if item.name != name:
                new_inventory.append(item)
        self.inventory = new_inventory
        msg.warning(f'You take {name} out of your inventory and toss it away.\nYour pack does feel a little lighter.')

    def __consume(self, item: Consumable):
        old_health = self.health
        self.health += item.value
        item.total -= 1
        if item.total == 0:
            for i, x in enumerate(self.inventory):
                if x.name == item.name:
                    self.inventory.pop(i)
                    msg.scroll_narrate(x.on_use)
                    break
        msg.success(f"Tasty. Your health has increased from {old_health} to {self.health}.")

    def __use_tool(self, tool: Tool, place: Place, target: str):
        for feature in place.features:
            if feature.name == target:
                break
        else:
            msg.warning(f"Huh? What do you want to use your {tool.name} on?")
            return
        if tool.name.lower() == feature.tool_name.lower():
            msg.success(feature.success_msg)
            tool.uses -= 1
            if feature.unlock:
                feature.unlock.accessible = True
        else:
            print('failed')

    def use(self, name: str, place: Place, target: str):
        for item in self.inventory:
            if item.name.lower() == name.lower():
                break
        else:
            msg.warning(f"Huh, you dig around in your pack and can't find anything like a {name}.")
        if isinstance(item, (Clue)):
            self.read(name)
        elif isinstance(item, Tool):
            self.__use_tool(item, place, target)

    def read(self, name: str):
        for item in self.inventory:
            if isinstance(item, Clue) and item.name.lower() == name.lower():
                msg.heading(f'Clue: "{item.name}"\t{item.form}')
                msg.narrate(item.content)
                msg.heading('end')
                return
        else:
            msg.warning(f"After thumbing through your clues and notes, you can't find a \"{name}\".")

    def wound(self, damage: int):
        self.health -= damage
        msg.danger(f'You are wounded for {damage} ponts.')
        msg.orange(f'Health: {self.health}')
        if self.health <= 0:
            self.lose()

    def lose(self):
        print('You lose, bummer. Finish this method')

    def status(self):
        msg.heading(f"{self.name}'s Status")
        if self.health >= 75:
            msg.success(f'Health: {self.health}')
        elif 30 < self.health <= 75:
            msg.warning(f'Health: {self.health}')
        else: msg.danger(f'Health: {self.health}')
        msg.plain(f'Score: {self.score}')
        msg.plain(f'Items in your inventory: {len(self.inventory)}')
        # plain(f'Strength: {self.strength}')
        msg.heading('end')
    
    def show_inventory(self):
        def sort_key(item):
            if isinstance(item, Consumable):
                return 1
            elif isinstance(item, Tool):
                return 3
            elif isinstance(item, Clue):
                return 5
            else:
                return 7
        sorted_inventory = sorted(self.inventory, key=sort_key)
        msg.heading(f"{self.name}'s Inventory")
        for item in sorted_inventory:
            msg.orange(item)
        msg.heading('end')

    def eat(self, item_name: str):
        for item in self.inventory:
            if item.name.lower() == item_name.lower() and isinstance(item, Consumable):
                self.__consume(item)
                break
        else:
            msg.warning(f"\nYou don't have a consummable named '{item_name}'")



from time import sleep
import re

from rich.console import Console


console = Console(color_system='256')

HELP = '''
Gameplay works by typing a command and a target.

The following commands are valid:
"take", "investigate", "use", "read", "go", "eat", "quit",
"save", "exit", "quit".

Make sure to include a target, for example:
"take book", "investigate nest", "read leather book", "go right".

The "use" command requires a tool [italic]and[/italic] a target,
for example: "[u]use key [/u] on [u]door lock[/u]"

Type "status" to see your character info and "hint"
to receive a help message if there is one.

Type "backpack" or "inventory" to list out all of your items.
'''.strip()

class Message:

    def __init__(self) -> None:
        self.text_scroll_delay = 0.01

    def danger(self, text: str):
        console.print(text, style='bold red')

    def warning(self, text: str):
        console.print(text, style='bold yellow')

    def narrate(self, text: str):
        console.print(text, style='italic')

    def dialogue(self, text: str):
        console.print(text, style='italic cyan')

    def success(self, text: str):
        console.print(text, style='bold green')

    def orange(self, text: str):
        console.print(text, style='orange1')

    def plain(self, text: str):
        console.print(text)

    def heading(self, text: str, color: str = 'magenta'):
        print()
        console.rule(f'[bold magenta]{text}', style=f'bold {color}')
        print()

    def scroll_text(self, text: list | str, style: str):
        print()
        if not isinstance(text, list):
            text: str = text.splitlines()
        spinner = console.status('', spinner='toggle2', spinner_style=style, speed=1.0, refresh_per_second=25)
        spinner.start()
        for line in text:
            words = []
            for word in line.split():
                characters = []
                for character in re.sub(r'\[[^\]]+\]', '', word): #strip out markup when rendering single characters
                    characters.append(character)
                    spinner.update(f'\t[{style}]{" ".join(words)} {"".join(characters)}')
                    sleep(self.text_scroll_delay)
                words.append(word)
            console.print(f'\t{line}', style=style)

        spinner.stop()
        print()

    def scroll_danger(self, text):
        self.self.scroll_text(text, 'bold red')

    def scroll_warning(self, text):
        self.scroll_text(text, 'bold yellow')

    def scroll_narrate(self, text):
        self.scroll_text(text, 'italic')

    def scroll_dialogue(self, text):
        self.scroll_text(text, 'italic cyan')

    def scroll_success(self, text):
        self.scroll_text(text, 'bold green')

    def scroll_orange(self, text):
        self.scroll_text(text, 'orange1')

    def scroll_plain(self, text):
        self.scroll_text(self, text, 'white on black')

    def scroll_danger(self, text):
        self.scroll_text(text, 'bold red')


    def help_text(self):
        console.rule('[bold]Help')
        self.plain(HELP)
        console.rule('[bold]End of Help')

msg = Message()



from dataclasses import dataclass

from .items import Tool, Clue, Consumable


@dataclass
class Node:
    name: str
    place_name: str
    travel_msg: str
    accessible: bool


@dataclass
class Feature:
    name: str
    interact_msg: str
    tool_name: str|None = None
    success_msg: str|None = None
    takeable: Tool|Clue|Consumable|None = None
    danger: int|None = None
    danger_msg: str|None = None
    unlock: Node|None = None
    # strength_required: int|None = None


@dataclass
class Place:
    '''nodes: places.Node'''
    name: str
    description: str
    features: list[Feature]
    nodes: list[Node]



from dataclasses import dataclass

@dataclass
class Consumable:
    name: str
    risk: int
    value: int
    description: str
    on_use: str
    total: int
    def __repr__(self) -> str:
        return f'Consumable: {self.name}\n\tTotal: {self.total}\tValue: {self.value}\tRisk: {self.risk}\n\tDescription: {self.description}'


@dataclass
class Tool:
    name: str
    risk: int
    uses: int
    description: str
    total: int
    def __repr__(self):
        return f'Tool: {self.name}\n\tTotal: {self.total}\t Uses left: {self.uses}\tRisk: {self.risk}\n\tDescription: {self.description}'


@dataclass
class Clue:
    name: str
    form: str # note, scroll, book, etc.
    content: str
    def __repr__(self):
        return f'Clue: {self.name}\tForm: {self.form}'



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyventure']

package_data = \
{'': ['*']}

install_requires = \
['rich>=12.4.4,<13.0.0']

setup_kwargs = {
    'name': 'pyventure',
    'version': '0.1.1',
    'description': 'A simple framework for creating text adventure games.',
    'long_description': '# Pyventure is a simple library for creating a text adventure with Python\nI created this library with and for my son for him practice coding,\nusing pip, and importing libraries.\n\n\n## Installation\n`python -m pip install pyventure`\n\n## Tutorial\nPyventure handles the game logic; all the user needs to do is\ncreate `Place` objects. Each `Place` object represents a location\non the game map. It is the game space which includes things with \nwhich the player can interact.\n\nA `Place` object is created by providing four parameters:\n1. a string `name` _required_\n2. a string `description` _required_\n3. a list of `Feature`s\n4. a list of `Node`s\n\n### Example of a playable game with two places to move between\n```python\nfrom pyventure.place import Place, Feature, Node\nfrom pyventure.items import Clue, Consumable, Tool\nfrom pyventure.game_loops import start\nfrom pyventure.message import msg\n\nLIVING_ROOM = \'Living Room\'\nKITCHEN = \'Kitchen\'\n\nliving_room = Place(\n    name=LIVING_ROOM,\n    description="There is a door on your [u]left[/u] and a pencil on the carpet.",\n    features=[\n        Feature(\n            name=\'pencil\',\n            interact_msg=\'It is a no. 2 Pencil\',\n            takeable=Tool(\n                name=\'pencil\',\n                risk=0,\n                uses=10,\n                description=\'it could stand to be sharpened\',\n                total=1\n            )\n        )\n    ],\n    nodes = [\n        Node(\n            name=\'left\',\n            place_name=KITCHEN,\n            travel_msg=\'You open the door and stop into the kitchen.\',\n            accessible=True\n        )\n    ]\n)\n\n\nkitchen = Place(\n    name=KITCHEN,\n    description="The floor is dirty. You\'re afraid of what is in the refridgerator.",\n    features=[],\n    nodes = [\n        Node(\n            name=\'living room\',\n            place_name=LIVING_ROOM,\n            travel_msg=\'You are back in the living room.\',\n            accessible=True\n        )\n    ]\n)\n\nall_places = {\n    LIVING_ROOM: living_room,\n    KITCHEN: kitchen\n}\n\n\nif __name__ == \'__main__\':\n    msg.narrate(\'Welcome to a Simple Game\')\n    start(new_game_msg=\'Name your character: \', all_places=all_places)\n\n```\n',
    'author': 'David Flood',
    'author_email': 'davidfloodii@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/d-flood/pyventure',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
